@@ -18,23 +18,156 @@ import json
 import math
 import os
 import random
+import re
 import socket
 import time
 
+import six
+
 import pyghmi.exceptions as pygexc
 import pyghmi.ipmi.private.util as util
+import pyghmi.media as media
 import pyghmi.redfish.oem.generic as generic
 import pyghmi.storage as storage
 from pyghmi.util.parse import parse_time
 import pyghmi.util.webclient as webclient
 
+numregex = re.compile('([0-9]+)')
+funtypes = {
+    0: 'RAID Controller',
+    1: 'Ethernet',
+    2: 'Fibre Channel',
+    3: 'Infiniband',
+    4: 'GPU',
+    10: 'NVMe Controller',
+    12: 'Fabric Controller',
+}
+
+
+def naturalize_string(key):
+    """Analyzes string in a human way to enable natural sort
+
+    :param key: string for the split
+    :returns: A structure that can be consumed by 'sorted'
+    """
+    return [int(text) if text.isdigit() else text.lower()
+            for text in re.split(numregex, key)]
+
+
+def natural_sort(iterable):
+    """Return a sort using natural sort if possible
+
+    :param iterable:
+    :return:
+    """
+    try:
+        return sorted(iterable, key=naturalize_string)
+    except TypeError:
+        # The natural sort attempt failed, fallback to ascii sort
+        return sorted(iterable)
+
 
 class OEMHandler(generic.OEMHandler):
+    logouturl = '/api/providers/logout'
+    bmcname = 'XCC'
+    ADP_URL = '/api/dataset/imm_adapters?params=pci_GetAdapters'
+    ADP_NAME = 'adapterName'
+    ADP_FUN = 'functions'
+    ADP_FU_URL = '/api/function/adapter_update?params=pci_GetAdapterListAndFW'
+    ADP_LABEL = 'connectorLabel'
+    ADP_SLOTNO = 'slotNo'
+    ADP_OOB = 'oobSupported'
+    ADP_PARTNUM = 'vpd_partNo'
+    ADP_SERIALNO = 'vpd_serialNo'
+    ADP_VENID = 'generic_vendorId'
+    ADP_SUBVENID = 'generic_subVendor'
+    ADP_DEVID = 'generic_devId'
+    ADP_SUBDEVID = 'generic_subDevId'
+    ADP_FRU = 'vpd_cardSKU'
+    BUSNO = 'generic_busNo'
+    PORTS = 'network_pPorts'
+    DEVNO = 'generic_devNo'
 
-    def __init__(self, sysinfo, sysurl, webclient, cache):
-        super(OEMHandler, self).__init__(sysinfo, sysurl, webclient, cache)
+    def __init__(self, sysinfo, sysurl, webclient, cache, gpool=None):
+        super(OEMHandler, self).__init__(sysinfo, sysurl, webclient, cache,
+                                         gpool)
         self._wc = None
         self.updating = False
+        self.datacache = {}
+
+    def get_cached_data(self, attribute, age=30):
+        try:
+            kv = self.datacache[attribute]
+            if kv[1] > util._monotonic_time() - age:
+                return kv[0]
+        except KeyError:
+            return None
+
+    def get_bmc_configuration(self):
+        settings = {}
+        passrules = self.wc.grab_json_response('/api/dataset/imm_users_global')
+        passrules = passrules.get('items', [{}])[0]
+        settings['password_reuse_count'] = {
+            'value': passrules.get('pass_min_resuse')}
+        settings['password_change_interval'] = {
+            'value': passrules.get('pass_change_interval')}
+        settings['password_expiration'] = {
+            'value': passrules.get('pass_expire_days')}
+        settings['password_login_failures'] = {
+            'value': passrules.get('max_login_failures')}
+        settings['password_complexity'] = {
+            'value': passrules.get('pass_complex_required')}
+        settings['password_min_length'] = {
+            'value': passrules.get('pass_min_length')}
+        settings['password_lockout_period'] = {
+            'value': passrules.get('lockout_period')}
+        presassert = self.wc.grab_json_response('/api/dataset/imm_rpp')
+        asserted = presassert.get('items', [{}])[0].get('rpp_Assert', None)
+        if asserted is not None:
+            settings['presence_assert'] = {
+                'value': 'Enable' if asserted else 'Disable'
+            }
+        return settings
+
+    rulemap = {
+        'password_change_interval': 'USER_GlobalMinPassChgInt',
+        'password_reuse_count': 'USER_GlobalMinPassReuseCycle',
+        'password_expiration': 'USER_GlobalPassExpPeriod',
+        'password_login_failures': 'USER_GlobalMaxLoginFailures',
+        'password_complexity': 'USER_GlobalPassComplexRequired',
+        'password_min_length': 'USER_GlobalMinPassLen',
+        'password_lockout_period': 'USER_GlobalLockoutPeriod',
+    }
+
+    def set_bmc_configuration(self, changeset):
+        ruleset = {}
+        for key in changeset:
+            if isinstance(changeset[key], six.string_types):
+                changeset[key] = {'value': changeset[key]}
+            currval = changeset[key].get('value', None)
+            if key.lower() in self.rulemap:
+                ruleset[self.rulemap[key.lower()]] = currval
+                if key.lower() == 'password_expiration':
+                    warntime = str(int(int(currval) * 0.08))
+                    ruleset['USER_GlobalPassExpWarningPeriod'] = warntime
+            elif 'presence_asserted'.startswith(key.lower()):
+                if 'enabled'.startswith(currval.lower()):
+                    self.wc.grab_json_response('/api/dataset',
+                                               {'IMM_RPPAssert': '0'})
+                    self.wc.grab_json_response('/api/dataset',
+                                               {'IMM_RPPAssert': '1'})
+                elif 'disabled'.startswith(currval.lower()):
+                    self.wc.grab_json_response('/api/dataset',
+                                               {'IMM_RPPAssert': '0'})
+                else:
+                    raise pygexc.InvalidParameterValue(
+                        '"{0}" is not a recognized value for {1}'.format(
+                            currval, key))
+            else:
+                raise pygexc.InvalidParameterValue(
+                    '{0} not a known setting'.format(key))
+        if ruleset:
+            self.wc.grab_json_response('/api/dataset', ruleset)
 
     def get_description(self):
         description = self._do_web_request('/DeviceDescription.json')
@@ -78,14 +211,14 @@ class OEMHandler(generic.OEMHandler):
                     bdata = {}
                     if 'versionStr' in firm and firm['versionStr']:
                         bdata['version'] = firm['versionStr']
-                    if ('releaseDate' in firm and
-                            firm['releaseDate'] and
-                            firm['releaseDate'] != 'N/A'):
+                    if ('releaseDate' in firm
+                            and firm['releaseDate']
+                            and firm['releaseDate'] != 'N/A'):
                         try:
                             bdata['date'] = parse_time(firm['releaseDate'])
                         except ValueError:
                             pass
-                    yield ('{0} {1}'.format(aname, fname), bdata)
+                    yield '{0} {1}'.format(aname, fname), bdata
 
     def _get_disk_firmware_single(self, diskent, prefix=''):
         bdata = {}
@@ -135,12 +268,17 @@ class OEMHandler(generic.OEMHandler):
 
     def get_storage_configuration(self, logout=True):
         rsp = self.wc.grab_json_response(
-            '/api/function/raid_alldevices?params=storage_GetAllDevices')
+            '/api/function/raid_alldevices?params=storage_GetAllDevices,0')
+        if not rsp:
+            rsp = self.wc.grab_json_response(
+                '/api/function/raid_alldevices?params=storage_GetAllDevices')
         standalonedisks = []
         pools = []
         for item in rsp.get('items', []):
             for cinfo in item['controllerInfo']:
-                cid = cinfo['id']
+                cid = '{0},{1},{2}'.format(
+                    cinfo['id'], cinfo.get('slotNo', -1), cinfo.get(
+                        'type', -1))
                 for pool in cinfo['pools']:
                     volumes = []
                     disks = []
@@ -232,9 +370,18 @@ class OEMHandler(generic.OEMHandler):
 
     def _raid_number_map(self, controller):
         themap = {}
+        cid = controller.split(',')
         rsp = self.wc.grab_json_response(
             '/api/function/raid_conf?'
-            'params=raidlink_GetDisksToConf,{0}'.format(controller))
+            'params=raidlink_GetDisksToConf,{0}'.format(cid[0]))
+        if rsp.get('return') == 22:  # New style firmware
+            if cid[2] == 2:
+                arg = '{0},{1}'.format(cid[1], cid[2])
+            else:
+                arg = '{0},{1}'.format(cid[0], cid[2])
+            arg = 'params=raidlink_GetDisksToConf,{0}'.format(arg)
+            rsp = self.wc.grab_json_response(
+                '/api/function/raid_conf?{0}'.format(arg))
         for lvl in rsp['items'][0]['supported_raidlvl']:
             mapdata = (lvl['rdlvl'], lvl['maxSpan'])
             raidname = lvl['rdlvlstr'].replace(' ', '').lower()
@@ -283,11 +430,18 @@ class OEMHandler(generic.OEMHandler):
             else:
                 hstr = ''
             drvstr = '|'.join([str(x.id[1]) for x in drives]) + '|'
+            ctl = controller.split(',')
             pth = '/api/function/raid_conf?params=raidlink_CheckConfisValid'
-            args = [pth, controller, rdlvl, spancount, drivesperspan, drvstr,
+            args = [pth, ctl[0], rdlvl, spancount, drivesperspan, drvstr,
                     hstr]
             url = ','.join([str(x) for x in args])
             rsp = self.wc.grab_json_response(url)
+            if rsp.get('return', -1) == 22:
+                args.append(ctl[1])
+                args = [pth, ctl[0], rdlvl, spancount, drivesperspan, drvstr,
+                        hstr, ctl[1]]
+                url = ','.join([str(x) for x in args])
+                rsp = self.wc.grab_json_response(url)
             if rsp['items'][0]['errcode'] == 16:
                 raise pygexc.InvalidParameterValue('Incorrect number of disks')
             elif rsp['items'][0]['errcode'] != 0:
@@ -310,8 +464,9 @@ class OEMHandler(generic.OEMHandler):
 
     def _create_array(self, pool):
         params = self._parse_array_spec(pool)
+        cid = params['controller'].split(',')[0]
         url = '/api/function/raid_conf?params=raidlink_GetDefaultVolProp'
-        args = (url, params['controller'], 0, params['drives'])
+        args = (url, cid, 0, params['drives'])
         props = self.wc.grab_json_response(','.join([str(x) for x in args]))
         props = props['items'][0]
         volumes = pool.volumes
@@ -336,18 +491,27 @@ class OEMHandler(generic.OEMHandler):
                     nameappend += 1
             else:
                 name = vol.name
-            if vol.stripsize:
+            if vol.stripsize is not None:
                 stripsize = int(math.log(vol.stripsize * 2, 2))
             else:
                 stripsize = props['stripsize']
+            if vol.read_policy is not None:
+                read_policy = vol.read_policy
+            else:
+                read_policy = props["cpra"]
+            if vol.write_policy is not None:
+                write_policy = vol.write_policy
+            else:
+                write_policy = props["cpwb"]
             strsize = 'remainder' if vol.size is None else str(vol.size)
             if strsize in ('all', '100%'):
                 volsize = params['capacity']
             elif strsize in ('remainder', 'rest'):
                 volsize = remainingcap
             elif strsize.endswith('%'):
-                volsize = int(params['capacity'] *
-                              float(strsize.replace('%', '')) / 100.0)
+                volsize = int(params['capacity']
+                              * float(strsize.replace('%', ''))
+                              / 100.0)
             else:
                 try:
                     volsize = int(strsize)
@@ -359,15 +523,26 @@ class OEMHandler(generic.OEMHandler):
                 raise pygexc.InvalidParameterValue(
                     'Requested sizes exceed available capacity')
             vols.append('{0};{1};{2};{3};{4};{5};{6};{7};{8};|'.format(
-                name, volsize, stripsize, props['cpwb'], props['cpra'],
+                name, volsize, stripsize, write_policy, read_policy,
                 props['cpio'], props['ap'], props['dcp'], props['initstate']))
         url = '/api/function'
+        cid = params['controller'].split(',')
+        cnum = cid[0]
         arglist = '{0},{1},{2},{3},{4},{5},'.format(
-            params['controller'], params['raidlevel'], params['spans'],
+            cnum, params['raidlevel'], params['spans'],
             params['perspan'], params['drives'], params['hotspares'])
         arglist += ''.join(vols)
         parms = {'raidlink_AddNewVolWithNaAsync': arglist}
         rsp = self.wc.grab_json_response(url, parms)
+        if rsp['return'] == 14:  # newer firmware
+            if cid[2] == 2:
+                cnum = cid[1]
+            arglist = '{0},{1},{2},{3},{4},{5},'.format(
+                cnum, params['raidlevel'], params['spans'],
+                params['perspan'], params['drives'], params['hotspares'])
+            arglist += ''.join(vols) + ',{0}'.format(cid[2])
+            parms = {'raidlink_AddNewVolWithNaAsync': arglist}
+            rsp = self.wc.grab_json_response(url, parms)
         if rsp['return'] != 0:
             raise Exception(
                 'Unexpected response to add volume command: ' + repr(rsp))
@@ -377,9 +552,18 @@ class OEMHandler(generic.OEMHandler):
         realcfg = self.get_storage_configuration(False)
         for pool in cfgspec.arrays:
             for volume in pool.volumes:
-                vid = '{0},{1}'.format(volume.id[1], volume.id[0])
+                cid = volume.id[0].split(',')
+                if cid[2] == 2:
+                    vid = '{0},{1},{2}'.format(volume.id[1], cid[1], cid[2])
+                else:
+                    vid = '{0},{1},{2}'.format(volume.id[1], cid[0], cid[2])
                 rsp = self.wc.grab_json_response(
                     '/api/function', {'raidlink_RemoveVolumeAsync': vid})
+                if rsp.get('return', -1) == 2:
+                    # older firmware
+                    vid = '{0},{1}'.format(volume.id[1], cid[0])
+                    rsp = self.wc.grab_json_response(
+                        '/api/function', {'raidlink_RemoveVolumeAsync': vid})
                 if rsp.get('return', -1) != 0:
                     raise Exception(
                         'Unexpected return to volume deletion: ' + repr(rsp))
@@ -403,8 +587,9 @@ class OEMHandler(generic.OEMHandler):
 
     @property
     def wc(self):
-        if (not self._wc or (self._wc.vintage and
-                             self._wc.vintage < util._monotonic_time() - 30)):
+        if (not self._wc or (self._wc.vintage
+                             and self._wc.vintage < util._monotonic_time()
+                             - 30)):
             self._wc = self.get_webclient()
         return self._wc
 
@@ -419,10 +604,13 @@ class OEMHandler(generic.OEMHandler):
             return None
         if not login:
             return wc
+        referer = 'https://xcc/'
         adata = json.dumps({'username': self.username,
                             'password': self.password
                             })
         headers = {'Connection': 'keep-alive',
+                   'Referer': referer,
+                   'Host': 'xcc',
                    'Content-Type': 'application/json'}
         wc.request('POST', '/api/login', adata, headers)
         rsp = wc.getresponse()
@@ -430,10 +618,29 @@ class OEMHandler(generic.OEMHandler):
             rspdata = json.loads(rsp.read())
             wc.set_header('Content-Type', 'application/json')
             wc.set_header('Authorization', 'Bearer ' + rspdata['access_token'])
+            wc.set_header('Referer', referer)
+            wc.set_header('Host', 'xcc')
             if '_csrf_token' in wc.cookies:
                 wc.set_header('X-XSRF-TOKEN', wc.cookies['_csrf_token'])
             return wc
 
+    def list_media(self, fishclient):
+        rt = self.wc.grab_json_response('/api/providers/rp_vm_remote_getdisk')
+        if 'items' in rt:
+            for mt in rt['items']:
+                url = mt['remotepath']
+                if url.startswith('//'):
+                    url = 'smb:' + url
+                elif (not url.startswith('http://')
+                      and not url.startswith('https://')):
+                    url = url.replace(':', '')
+                    url = 'nfs://' + url
+                yield media.Media(mt['filename'], url)
+        rt = self.wc.grab_json_response('/api/providers/rp_rdoc_imagelist')
+        if 'items' in rt:
+            for mt in rt['items']:
+                yield media.Media(mt['filename'])
+
     def attach_remote_media(self, url, user, password, vmurls):
         for vmurl in vmurls:
             if 'EXT' not in vmurl:
@@ -479,10 +686,14 @@ class OEMHandler(generic.OEMHandler):
         if rsp.get('return', -1) != 0:
             errmsg = repr(rsp) if rsp else self.wc.lastjsonerror
             raise Exception('Unrecognized return: ' + errmsg)
-        rsp = self.wc.grab_json_response('/api/providers/rp_rdoc_getfiles')
-        if 'items' not in rsp or len(rsp['items']) == 0:
-            raise Exception(
-                'Image upload was not accepted, it may be too large')
+        ready = False
+        while not ready:
+            time.sleep(3)
+            rsp = self.wc.grab_json_response('/api/providers/rp_rdoc_getfiles')
+            if 'items' not in rsp or len(rsp['items']) == 0:
+                raise Exception(
+                    'Image upload was not accepted, it may be too large')
+            ready = rsp['items'][0]['size'] != 0
         self._refresh_token()
         rsp = self.wc.grab_json_response('/api/providers/rp_rdoc_mountall',
                                          {})
@@ -533,8 +744,8 @@ class OEMHandler(generic.OEMHandler):
                 progress({'phase': 'upload',
                           'progress': 100.0 * rsp['received'] / rsp['size']})
             elif rsp['state'] != 'done':
-                if (rsp.get('status', None) == 413 or
-                        uploadthread.rspstatus == 413):
+                if (rsp.get('status', None) == 413
+                        or uploadthread.rspstatus == 413):
                     raise Exception('File is larger than supported')
                 raise Exception('Unexpected result:' + repr(rsp))
             uploadstate = rsp['state']
@@ -662,10 +873,10 @@ class OEMHandler(generic.OEMHandler):
 
         if rsp.get('return', -1) != 0:
             errmsg = repr(rsp) if rsp else self.wc.lastjsonerror
-            raise Exception('Unexpected result starting update: ' +
-                            errmsg)
+            raise Exception('Unexpected result starting update: %s' % errmsg)
         complete = False
         while not complete:
+            self._refresh_token()
             time.sleep(3)
             rsp = self.wc.grab_json_response(
                 '/api/dataset/imm_firmware_progress')
@@ -677,15 +888,15 @@ class OEMHandler(generic.OEMHandler):
             if rsp['items'][0]['action_state'] == 'Complete OK':
                 complete = True
                 if rsp['items'][0]['action_status'] != 0:
-                    raise Exception('Unexpected failure: ' + repr(rsp))
+                    raise Exception('Unexpected failure: %s' % repr(rsp))
                 break
-            if (rsp['items'][0]['action_state'] == 'In Progress' and
-                    rsp['items'][0]['action_status'] == 2):
+            if (rsp['items'][0]['action_state'] == 'In Progress'
+                    and rsp['items'][0]['action_status'] == 2):
                 raise Exception('Unexpected failure: ' + repr(rsp))
             if rsp['items'][0]['action_state'] != 'In Progress':
                 raise Exception(
                     'Unknown condition waiting for '
-                    'firmware update: ' + repr(rsp))
+                    'firmware update: %s' % repr(rsp))
         if bank == 'backup':
             return 'complete'
         return 'pending'
@@ -716,6 +927,8 @@ class OEMHandler(generic.OEMHandler):
                 progress({'phase': 'download',
                           'progress': 100 * self.wc.get_download_progress()})
             self._refresh_token()
+        if fd.exc:
+            raise fd.exc
         if progress:
             progress({'phase': 'complete'})
         return savefile
@@ -761,7 +974,7 @@ class OEMHandler(generic.OEMHandler):
                     self._refresh_token()
                 yield savefile
 
-    def apply_license(self, filename, progress=None):
+    def apply_license(self, filename, progress=None, data=None):
         license_errors = {
             310: "License is for a different model of system",
             311: "License is for a different system serial number",
@@ -769,7 +982,8 @@ class OEMHandler(generic.OEMHandler):
             313: "License is expired",
             314: "License usage limit reached",
         }
-        uploadthread = webclient.FileUploader(self.wc, '/upload', filename)
+        uploadthread = webclient.FileUploader(self.wc, '/upload', filename,
+                                              data=data)
         uploadthread.start()
         uploadthread.join()
         rsp = json.loads(uploadthread.rsp)
@@ -795,3 +1009,209 @@ class OEMHandler(generic.OEMHandler):
                     return 0
                 else:
                     return days
+
+    def get_inventory_descriptions(self, withids=False):
+        hwmap = self.hardware_inventory_map()
+        yield "System"
+        for key in natural_sort(hwmap):
+            yield key
+
+    def get_inventory_of_component(self, compname):
+        if compname.lower() == 'system':
+            sysinfo = {
+                'UUID': self._varsysinfo.get('UUID', ''),
+                'Serial Number': self._varsysinfo.get('SerialNumber', ''),
+                'Manufacturer': self._varsysinfo.get('Manufacturer', ''),
+                'Product Name': self._varsysinfo.get('Model', ''),
+                'Model': self._varsysinfo.get(
+                    'SKU', self._varsysinfo.get('PartNumber', '')),
+            }
+            return sysinfo
+        hwmap = self.hardware_inventory_map()
+        try:
+            return hwmap[compname]
+        except KeyError:
+
+            return None
+
+    def get_inventory(self, withids=False):
+        sysinfo = {
+            'UUID': self._varsysinfo.get('UUID', ''),
+            'Serial Number': self._varsysinfo.get('SerialNumber', ''),
+            'Manufacturer': self._varsysinfo.get('Manufacturer', ''),
+            'Product Name': self._varsysinfo.get('Model', ''),
+            'Model': self._varsysinfo.get(
+                'SKU', self._varsysinfo.get('PartNumber', '')),
+        }
+        yield ('System', sysinfo)
+        for cpuinv in self._get_cpu_inventory():
+            yield cpuinv
+        for meminv in self._get_mem_inventory():
+            yield meminv
+        hwmap = self.hardware_inventory_map()
+        for key in natural_sort(hwmap):
+            yield (key, hwmap[key])
+
+    def hardware_inventory_map(self):
+        hwmap = self.get_cached_data('lenovo_cached_hwmap')
+        if hwmap:
+            return hwmap
+        hwmap = {}
+        for disk in self.disk_inventory(mode=1):  # hardware mode
+            hwmap[disk[0]] = disk[1]
+        adapterdata = self.get_cached_data('lenovo_cached_adapters')
+        if not adapterdata:
+            if self.updating:
+                raise pygexc.TemporaryError(
+                    'Cannot read extended inventory during firmware update')
+            if self.wc:
+                adapterdata = self.wc.grab_json_response(self.ADP_URL)
+                if adapterdata:
+                    self.datacache['lenovo_cached_adapters'] = (
+                        adapterdata, util._monotonic_time())
+        if adapterdata and 'items' in adapterdata:
+            anames = {}
+            for adata in adapterdata['items']:
+                skipadapter = False
+                clabel = adata[self.ADP_LABEL]
+                if clabel == 'Unknown':
+                    continue
+                if clabel != 'Onboard':
+                    aslot = adata[self.ADP_SLOTNO]
+                    if clabel == 'ML2':
+                        clabel = 'ML2 (Slot {0})'.format(aslot)
+                    else:
+                        clabel = 'Slot {0}'.format(aslot)
+                aname = adata[self.ADP_NAME]
+                bdata = {'location': clabel, 'name': aname}
+                if aname in anames:
+                    anames[aname] += 1
+                    aname = '{0} {1}'.format(aname, anames[aname])
+                else:
+                    anames[aname] = 1
+                for fundata in adata[self.ADP_FUN]:
+                    bdata['pcislot'] = '{0:02x}:{1:02x}'.format(
+                        fundata[self.BUSNO], fundata[self.DEVNO])
+                    serialdata = fundata.get(self.ADP_SERIALNO, None)
+                    if (serialdata and serialdata != 'N/A'
+                            and '---' not in serialdata):
+                        bdata['serial'] = serialdata
+                    partnum = fundata.get(self.ADP_PARTNUM, None)
+                    if partnum and partnum != 'N/A':
+                        bdata['Part Number'] = partnum
+                    cardtype = funtypes.get(fundata.get('funType', None),
+                                            None)
+                    if cardtype is not None:
+                        bdata['Type'] = cardtype
+                    venid = fundata.get(self.ADP_VENID, None)
+                    if venid is not None:
+                        bdata['PCI Vendor ID'] = '{0:04x}'.format(venid)
+                    devid = fundata.get(self.ADP_DEVID, None)
+                    if devid is not None:
+                        bdata['PCI Device ID'] = '{0:04x}'.format(devid)
+                    venid = fundata.get(self.ADP_SUBVENID, None)
+                    if venid is not None:
+                        bdata['PCI Subsystem Vendor ID'] = '{0:04x}'.format(
+                            venid)
+                    devid = fundata.get(self.ADP_SUBDEVID, None)
+                    if devid is not None:
+                        bdata['PCI Subsystem Device ID'] = '{0:04x}'.format(
+                            devid)
+                    fruno = fundata.get(self.ADP_FRU, None)
+                    if fruno is not None:
+                        bdata['FRU Number'] = fruno
+                    if self.PORTS in fundata:
+                        for portinfo in fundata[self.PORTS]:
+                            for lp in portinfo['logicalPorts']:
+                                ma = lp['networkAddr']
+                                ma = ':'.join(
+                                    [ma[i:i + 2] for i in range(
+                                        0, len(ma), 2)]).lower()
+                                bdata['MAC Address {0}'.format(
+                                    portinfo['portIndex'])] = ma
+                    elif clabel == 'Onboard':  # skip the various non-nic
+                        skipadapter = True
+                if not skipadapter:
+                    hwmap[aname] = bdata
+            self.datacache['lenovo_cached_hwmap'] = (hwmap,
+                                                     util._monotonic_time())
+        # self.weblogout()
+        return hwmap
+
+    def disk_inventory(self, mode=0):
+        # mode 0 is firmware, 1 is hardware
+        storagedata = self.get_cached_data('lenovo_cached_storage')
+        if not storagedata:
+            if self.wc:
+                storagedata = self.wc.grab_json_response(
+                    '/api/function/raid_alldevices?params=storage_GetAllDisks')
+                if storagedata:
+                    self.datacache['lenovo_cached_storage'] = (
+                        storagedata, util._monotonic_time())
+        if storagedata and 'items' in storagedata:
+            for adp in storagedata['items']:
+                for diskent in adp.get('disks', ()):
+                    if mode == 0:
+                        yield self.get_disk_firmware(diskent)
+                    elif mode == 1:
+                        yield self.get_disk_hardware(diskent)
+                for diskent in adp.get('aimDisks', ()):
+                    if mode == 0:
+                        yield self.get_disk_firmware(diskent)
+                    elif mode == 1:
+                        yield self.get_disk_hardware(diskent)
+                if mode == 1:
+                    bdata = {'Description': 'Unmanaged Disk'}
+                    if adp.get('m2Type', -1) == 2:
+                        yield 'M.2 Disk', bdata
+                    for umd in adp.get('unmanagedDisks', []):
+                        yield 'Disk {0}'.format(umd['slotNo']), bdata
+
+    def _get_cpu_inventory(self):
+        procdata = self.get_cached_data('lenovo_cached_proc')
+        if not procdata:
+            if self.wc:
+                procdata = self.wc.grab_json_response(
+                    '/api/dataset/imm_processors')
+                if procdata:
+                    self.datacache['lenovo_cached_proc'] = (
+                        procdata, util._monotonic_time())
+        if procdata:
+            for proc in procdata.get('items', [{}])[0].get('processors', []):
+                procinfo = {
+                    'Model': proc['processors_cpu_model']
+                }
+                yield ('Processor {0}'.format(proc['processors_name']),
+                       procinfo)
+
+    def _get_mem_inventory(self):
+        memdata = self.get_cached_data('lenovo_cached_memory')
+        if not memdata:
+            if self.wc:
+                memdata = self.wc.grab_json_response(
+                    '/api/dataset/imm_memory')
+                if memdata:
+                    self.datacache['lenovo_cached_memory'] = (
+                        memdata, util._monotonic_time())
+        if memdata:
+            for dimm in memdata.get('items', [{}])[0].get('memory', []):
+                memdata = {}
+                memdata['speed'] = dimm['memory_mem_speed'] * 8 // 100 * 100
+                memdata['module_type'] = 'RDIMM'
+                memdata['capacity_mb'] = dimm['memory_capacity'] * 1024
+                memdata['manufacturer'] = dimm['memory_manufacturer']
+                memdata['memory_type'] = dimm['memory_type']
+                memdata['model'] = dimm['memory_part_number'].rstrip()
+                memdata['serial'] = dimm['memory_serial_number']
+                yield (dimm['memory_name'], memdata)
+
+    def get_disk_hardware(self, diskent, prefix=''):
+        bdata = {}
+        if not prefix and diskent.get('location', '').startswith('M.2'):
+            prefix = 'M.2-'
+        diskname = 'Disk {1}{0}'.format(diskent['slotNo'], prefix)
+        bdata['Model'] = diskent['productName'].rstrip()
+        bdata['Serial Number'] = diskent['serialNo'].rstrip()
+        bdata['FRU Number'] = diskent['fruPartNo'].rstrip()
+        bdata['Description'] = diskent['type'].rstrip()
+        return (diskname, bdata)
